package org.intellij.ipp.fqnnames;

import com.intellij.psi.*;
import com.intellij.psi.util.ClassUtil;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.util.IncorrectOperationException;
import com.siyeh.ipp.base.Intention;
import com.siyeh.ipp.base.PsiElementPredicate;
import com.siyeh.ipp.psiutils.ImportUtils;
import org.intellij.ipp.util.FQNUtil;
import org.intellij.ipp.util.FQNBundle;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

/**
 * Copyright 2006 Cojocar Dan
 * <p/>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p/>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p/>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 * <b>New Replace Qualified Name With Import</b> is replacing
 * <b>Replace Qualified Name With Import</b> intention from IntentionPowerPack.<br>
 * Compared to the IntentionPowerPack implementation this is fixing some bugs and is adding new functionalities:
 * <ul>
 *  <li>will be able to perform the intention only if this is possible.
 *  <li>will be able to apply the intention on expressions too, not only on type definitions.
 *  <li>will add an import statement only if it is needed.
 * </ul>
 * 
 * User: dan
 * Date: Oct 24, 2006
 * Time: 2:52:02 PM
 * $Id$
 */
public class FQNIntention extends Intention {

    /**
     * Associate the predicate
     * @see com.siyeh.ipp.base.Intention#getElementPredicate()
     * @return our predicate
     */
    @NotNull
    public PsiElementPredicate getElementPredicate() {
        return new FQNPredicate();
    }

    /**
     * Intention mechanism @see com.siyeh.ipp.base.Intention#processIntention
     * @param element the element that we will inspect
     * @throws IncorrectOperationException
     */
    public void processIntention(@NotNull PsiElement element)
            throws IncorrectOperationException {
        PsiJavaCodeReferenceElement reference = FQNUtil.getReferencedName(element);

        final PsiJavaFile javaFile =PsiTreeUtil.getParentOfType(reference, PsiJavaFile.class);
        if (javaFile == null) {
            return;
        }
        final PsiElement target = reference.resolve();
        if (!(target instanceof PsiClass)) {
            return;
        }
        final PsiClass aClass = (PsiClass) target;
        final String qualifiedName = aClass.getQualifiedName();
        if (qualifiedName == null) {
            return;
        }
        if (!FQNUtil.nameCanBeImported(qualifiedName, javaFile))
            return;

        final PsiImportList importList = javaFile.getImportList();
        if (importList == null) {
            return;
        }
        @NonNls final String packageName = ClassUtil.extractPackageName(qualifiedName);
        if (packageName.equals("java.lang")) {
            if (ImportUtils.hasOnDemandImportConflict(qualifiedName, javaFile)) {
                addImport(importList, aClass);
            }
        } else {
            if (importList.findSingleClassImportStatement(qualifiedName) == null &&
                    !FQNUtil.isInPackage(target,javaFile)) {
                addImport(importList, aClass);
            }
        }
        final PsiElement qualifier = reference.getQualifier();
        if (qualifier == null) {
            return;
        }
        qualifier.delete();
    }

    /**
     * Add import statement
     * @param importList the import list
     * @param aClass the class that will be imported
     * @throws IncorrectOperationException
     */
    private static void addImport(PsiImportList importList, PsiClass aClass)
            throws IncorrectOperationException {
        final PsiManager manager = importList.getManager();
        final PsiElementFactory elementFactory =
                manager.getElementFactory();
        final PsiImportStatement importStatement =
                elementFactory.createImportStatement(aClass);
        importList.add(importStatement);
    }

    /**
     * Create a string representation used by @link getText and @link getFamilyName
     * @return the prefix
     */
    private String getPrefix() {
        final Class<? extends Intention> aClass = getClass();
        final String name = aClass.getSimpleName();
        final StringBuilder buffer = new StringBuilder(name.length() + 10);
        buffer.append(Character.toLowerCase(name.charAt(0)));
        for (int i = 1; i < name.length(); i++){
            final char c = name.charAt(i);
            if (Character.isUpperCase(c)){
                buffer.append('.');
                buffer.append(Character.toLowerCase(c));
            } else {
                buffer.append(c);
            }
        }
        return buffer.toString();
    }

    /**
     * @see com.siyeh.ipp.base.Intention#getText()
     * @return the text for the intention
     */
    @NotNull
    public String getText() {
        return FQNBundle.message(getPrefix() + ".name");
    }

    /**
     * @see com.siyeh.ipp.base.Intention#getFamilyName()
     * @return the family text name for this intention
     */
    @NotNull
    public String getFamilyName() {
        return FQNBundle.message(getPrefix() + ".family.name");
    }

}
