package org.intellij.ipp.util;

import com.intellij.psi.*;
import com.intellij.psi.util.ClassUtil;
import org.jetbrains.annotations.NotNull;

/**
 * Copyright 2006 Cojocar Dan
 * <p/>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p/>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p/>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * A modified copy of @see com.siyeh.ipp.psiutils.ImportUtils
 *
 * User: dan
 * Date: Oct 24, 2006
 * Time: 4:24:41 PM
 * $Id$
 */
public class FQNUtil {

    public static boolean nameCanBeImported(@NotNull String fqName,
                                            @NotNull PsiJavaFile file) {
        return !containsConflictingClass(fqName, file) &&
                !containsConflictingClassName(fqName, file) &&
                !hasExactImportConflict(fqName, file);
    }

    private static boolean hasExactImportConflict(String fqName,
                                                  PsiJavaFile file) {
        final PsiImportList imports = file.getImportList();
        if (imports == null) {
            return false;
        }
        final PsiImportStatement[] importStatements = imports
                .getImportStatements();
        final int lastDotIndex = fqName.lastIndexOf((int) '.');
        final String shortName = fqName.substring(lastDotIndex + 1);
        final String dottedShortName = '.' + shortName;
        for (final PsiImportStatement importStatement : importStatements) {
            if (!importStatement.isOnDemand()) {
                final String importName = importStatement.getQualifiedName();
                if (importName == null) {
                    return false;
                }
                if (!importName.equals(fqName)) {
                    if (importName.endsWith(dottedShortName)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    private static boolean containsConflictingClass(String fqName,
                                                    PsiJavaFile file) {
        final PsiClass[] classes = file.getClasses();
        for (PsiClass aClass : classes) {
            if (containsConflictingInnerClass(fqName, aClass)) {
                return true;
            }
        }
        final ClassReferenceVisitor visitor =
                new ClassReferenceVisitor(fqName);
        file.accept(visitor);
        return visitor.isReferenceFound();
    }


    private static boolean containsConflictingClassName(String fqName,
                                                        PsiJavaFile file) {
        final int lastDotIndex = fqName.lastIndexOf((int) '.');
        final String shortName = fqName.substring(lastDotIndex + 1);
        final PsiClass[] classes = file.getClasses();
        for (PsiClass aClass : classes) {
            if (shortName.equals(aClass.getName()) && !fqName.equals(aClass.getQualifiedName())) {
                return true;
            }
        }
        return false;
    }


    public static boolean isInPackage(PsiElement target, PsiJavaFile javaFile) {
        if (target == null || javaFile == null || javaFile.getPackageStatement() == null)
            return false;
        PsiPackageStatement packageStatement = javaFile.getPackageStatement();
        PsiJavaCodeReferenceElement reference;
        if (packageStatement != null) {
            reference = packageStatement.getPackageReference();
            PsiPackage psiPackage = (PsiPackage) reference.resolve();
            if (psiPackage != null) {
                return javaFile.getManager().isInPackage(target, psiPackage);
            }
        }
        return false;
    }

    private static boolean containsConflictingInnerClass(String fqName,
                                                         PsiClass aClass) {
        final String shortName = ClassUtil.extractClassName(fqName);
        if (shortName.equals(aClass.getName())) {
            if (!fqName.equals(aClass.getQualifiedName())) {
                return true;
            }
        }
        final PsiClass[] classes = aClass.getInnerClasses();
        for (PsiClass innerClass : classes) {
            if (containsConflictingInnerClass(fqName, innerClass)) {
                return true;
            }
        }
        return false;
    }

    public static PsiJavaCodeReferenceElement getReferencedName(PsiElement element) {
        PsiJavaCodeReferenceElement reference =
                (PsiJavaCodeReferenceElement) element;
        while (reference.getParent() instanceof PsiJavaCodeReferenceElement && !(reference.resolve() instanceof PsiClass)) {
            reference = (PsiJavaCodeReferenceElement) reference.getParent();
        }
        if (!(reference.resolve() instanceof PsiClass)) {
            reference = (PsiJavaCodeReferenceElement) element;
            while (reference != null && reference.getFirstChild() instanceof PsiJavaCodeReferenceElement
                    && !(reference.resolve() instanceof PsiClass)) {
                reference = (PsiJavaCodeReferenceElement) reference.getFirstChild();
            }
        }
        return reference;
    }

    private static class ClassReferenceVisitor
            extends PsiRecursiveElementVisitor {

        private final String m_name;
        private final String fullyQualifiedName;
        private boolean m_referenceFound = false;

        private ClassReferenceVisitor(String fullyQualifiedName) {
            super();
            m_name = ClassUtil.extractClassName(fullyQualifiedName);
            this.fullyQualifiedName = fullyQualifiedName;
        }

        public void visitReferenceExpression(
                PsiReferenceExpression expression) {
            super.visitReferenceExpression(expression);
            if (m_referenceFound) {
                return;
            }
            final String text = expression.getText();
            if (text.indexOf((int) '.') >= 0 || !m_name.equals(text)) {
                return;
            }
            final PsiElement element = expression.resolve();
            if (!(element instanceof PsiClass)
                    || element instanceof PsiTypeParameter) {
                return;
            }
            final PsiClass aClass = (PsiClass) element;
            final String testClassName = aClass.getName();
            final String testClassQualifiedName = aClass.getQualifiedName();
            if (testClassQualifiedName == null || testClassName == null) {
                return;
            }
            if (testClassQualifiedName.equals(fullyQualifiedName) ||
                    !testClassName.equals(m_name)) {
                return;
            }
            m_referenceFound = true;
        }

        public boolean isReferenceFound() {
            return m_referenceFound;
        }
    }

}
